@@ -106,6 +106,7 @@ struct proxy_remote {
 
 #define PROXYPASS_NOCANON 0x01
 #define PROXYPASS_INTERPOLATE 0x02
+#define PROXYPASS_NOQUERY 0x04
 struct proxy_alias {
     const char  *real;
     const char  *fake;
@@ -423,6 +424,7 @@ typedef struct {
     unsigned int    need_reset:1;
     unsigned int    vhosted:1;
     unsigned int    inactive:1;
+    unsigned int    forcerecovery:1;
 } proxy_balancer_shared;
 
 #define ALIGNED_PROXY_BALANCER_SHARED_SIZE (APR_ALIGN_DEFAULT(sizeof(proxy_balancer_shared)))
